# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import traceback
from dataclasses import dataclass
from threading import Thread
from typing import Dict

from datahub_actions.pipeline.pipeline import Pipeline, PipelineException

logger = logging.getLogger(__name__)


@dataclass
class PipelineSpec:
    # The pipeline name
    name: str

    # The pipeline
    pipeline: Pipeline

    # The thread which is executing the pipeline.
    thread: Thread


# Run a pipeline in blocking fashion
# TODO: Exit process on failure of single pipeline.
def run_pipeline(pipeline: Pipeline) -> None:
    try:
        pipeline.run()
    except PipelineException:
        logger.error(
            f"Caught exception while running pipeline with name {pipeline.name}: {traceback.format_exc(limit=3)}"
        )
        pipeline.stop()
    logger.debug(f"Thread for pipeline with name {pipeline.name} has stopped.")


# A manager of multiple Action Pipelines.
# This class manages 1 thread per pipeline registered.
class PipelineManager:
    # A catalog of all the currently executing Action Pipelines.
    pipeline_registry: Dict[str, PipelineSpec] = {}

    def __init__(self) -> None:
        pass

    # Start a new Action Pipeline.
    def start_pipeline(self, name: str, pipeline: Pipeline) -> None:
        logger.debug(f"Attempting to start pipeline with name {name}...")
        if name not in self.pipeline_registry:
            thread = Thread(target=run_pipeline, args=([pipeline]))
            thread.start()
            spec = PipelineSpec(name, pipeline, thread)
            self.pipeline_registry[name] = spec
            logger.debug(f"Started pipeline with name {name}.")
        else:
            raise Exception(f"Pipeline with name {name} is already running.")

    # Stop a running Action Pipeline.
    def stop_pipeline(self, name: str) -> None:
        logger.debug(f"Attempting to stop pipeline with name {name}...")
        if name in self.pipeline_registry:
            # First, stop the pipeline.
            try:
                pipeline_spec = self.pipeline_registry[name]
                pipeline_spec.pipeline.stop()
                pipeline_spec.thread.join()  # Wait for the pipeline thread to terminate.
                logger.info(f"Actions Pipeline with name '{name}' has been stopped.")
                pipeline_spec.pipeline.stats().pretty_print_summary(
                    name
                )  # Print the pipeline's statistics.
                del self.pipeline_registry[name]
            except Exception as e:
                # Failed to stop a pipeline - this is a critical issue, we should avoid starting another action of the same type
                # until this pipeline is confirmed killed.
                logger.error(
                    f"Caught exception while attempting to stop pipeline with name {name}: {traceback.format_exc(limit=3)}"
                )
                raise Exception(
                    f"Caught exception while attempting to stop pipeline with name {name}."
                ) from e
        else:
            raise Exception(f"No pipeline with name {name} found.")

    # Stop all running pipelines.
    def stop_all(self) -> None:
        logger.debug("Attempting to stop all running pipelines...")
        # Stop each running pipeline.
        names = list(self.pipeline_registry.keys()).copy()
        for name in names:
            self.stop_pipeline(name)
        logger.debug("Successfully stop all running pipelines.")



import datetime
import logging
from dataclasses import dataclass
from typing import Dict, Optional

from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.schema_classes import EntityChangeEventClass as EntityChangeEvent
from datahub.utilities.urns.urn import Urn
from datahub_actions.utils.name_resolver import (
    get_entity_name_from_urn,
    get_entity_qualifier_from_urn,
)

logger = logging.getLogger(__name__)


def pretty_any_text(text: str, channel: str) -> str:
    if text.startswith("http:"):
        return make_url_with_title(text, text, channel)
    else:
        return text


def make_url_with_title(title: str, url: str, channel: str) -> str:
    if channel == "slack":
        # slack uses mrkdwn format
        return f"<{url}|{title}>"
    else:
        return f"[{title}]({url})"


def make_bold(text: str, channel: str) -> str:
    if not text:
        return text
    if channel == "slack":
        return f"*{text}*"
    else:
        return f"**{text}**"


@dataclass
class StructuredMessage:
    title: str
    properties: Dict[str, str]
    text: Optional[str]


def get_welcome_message(datahub_home_url: str) -> StructuredMessage:
    hostname = "unknown-host"
    try:
        import os

        hostname = os.uname()[1]
    except Exception as e:
        logger.warning(f"Failed to acquire hostname with {e}")
        pass

    current_time: str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    current_timezone: str = str(datetime.datetime.now().astimezone().tzinfo)
    return StructuredMessage(
        title="DataHub Bot 🌟",
        properties={
            "Home 🏠": datahub_home_url,
            "Host 🖥️": hostname,
            "Started ⏱️": f"{current_time} ({current_timezone})",
        },
        text=f"👀 I'll be watching for interesting events on DataHub at <{datahub_home_url}> and keep you updated when anything changes. ⚡",
    )


def get_message_from_entity_change_event(
    event: EntityChangeEvent,
    datahub_base_url: str,
    datahub_graph: Optional[DataHubGraph],
    channel: str,
) -> Optional[str]:
    datahub_base_url = datahub_base_url.rstrip("/")
    actor_name = get_entity_name_from_urn(event.auditStamp.actor, datahub_graph)
    operation = {
        "ADD": "added",
        "UPDATE": "updated",
        "MODIFY": "updated",
        "REMOVE": "removed",
        "CREATE": "created",
        "REINSTATE": "reinstated",
    }.get(event.operation, event.operation.lower())
    if event.modifier:
        try:
            modifier_name = get_entity_name_from_urn(event.modifier, datahub_graph)
        except Exception:
            modifier_name = ""
    else:
        modifier_name = ""
    category = event.category.lower()

    entity_specialized_type = get_entity_qualifier_from_urn(
        event.entityUrn, datahub_graph
    )
    entity_name = get_entity_name_from_urn(event.entityUrn, datahub_graph)
    # special handling for entity type schemaField
    if event.entityType == "schemaField":
        schema_field_urn = Urn.from_string(event.entityUrn)
        parent_entity_urn = Urn.from_string(schema_field_urn.get_entity_id()[0])
        parent_entity_name = get_entity_name_from_urn(
            str(parent_entity_urn), datahub_graph
        )
        parent_specialized_type = get_entity_qualifier_from_urn(
            str(parent_entity_urn), datahub_graph
        )
        parent_entity_url = f"{datahub_base_url}/{parent_entity_urn.get_type()}/{parent_entity_urn}/Schema?schemaFilter={entity_name}"
        entity_message_trailer = f"{entity_name} of {parent_specialized_type} {make_url_with_title(title=parent_entity_name, url=parent_entity_url, channel=channel)}"
    elif event.entityType == "dataFlow":
        entity_url = f"{datahub_base_url}/pipelines/{event.entityUrn}"
        entity_message_trailer = make_url_with_title(
            title=entity_name, url=entity_url, channel=channel
        )
    elif event.entityType == "dataJob":
        entity_url = f"{datahub_base_url}/tasks/{event.entityUrn}"
        entity_message_trailer = make_url_with_title(
            title=entity_name, url=entity_url, channel=channel
        )
    else:
        entity_url = f"{datahub_base_url}/{event.entityType}/{event.entityUrn}"
        entity_message_trailer = make_url_with_title(
            title=entity_name, url=entity_url, channel=channel
        )

    if category == "lifecycle":
        message = f">✏️ {make_bold(actor_name, channel)} has {operation} {entity_specialized_type} {entity_message_trailer}."
    elif category == "technical_schema":
        if event.modifier and event.modifier.startswith("urn:li:schemaField"):
            message = f">✏️ {make_bold(actor_name, channel)} has {operation} field {make_bold(modifier_name, channel)} in schema for {entity_specialized_type} {entity_message_trailer}."
        else:
            message = f">✏️ {make_bold(actor_name, channel)} has {operation} {make_bold(modifier_name, channel)} schema for {entity_specialized_type} {entity_message_trailer}."
    else:
        message = f">✏️ {make_bold(actor_name, channel)} has {operation} {category} {make_bold(modifier_name, channel)} for {entity_specialized_type} {entity_message_trailer}."
    return message



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
from time import time
from typing import Dict

import click

from datahub_actions.action.action_stats import ActionStats
from datahub_actions.pipeline.pipeline_util import get_transformer_name
from datahub_actions.transform.transformer import Transformer
from datahub_actions.transform.transformer_stats import TransformerStats


# Class that stores running statistics for a single Actions Pipeline.
class PipelineStats:
    # Timestamp in milliseconds when the pipeline was launched.
    started_at: int

    # Number of events that failed processing even after retry.
    failed_event_count: int = 0

    # Number of events that failed when "ack" was invoked.
    failed_ack_count: int = 0

    # Top-level number of succeeded processing executions.
    success_count: int = 0

    # Transformer Stats
    transformer_stats: Dict[str, TransformerStats] = {}

    # Action Stats
    action_stats: ActionStats = ActionStats()

    def mark_start(self) -> None:
        self.started_at = int(time() * 1000)

    def increment_failed_event_count(self) -> None:
        self.failed_event_count = self.failed_event_count + 1

    def increment_failed_ack_count(self) -> None:
        self.failed_ack_count = self.failed_ack_count + 1

    def increment_success_count(self) -> None:
        self.success_count = self.success_count + 1

    def increment_transformer_exception_count(self, transformer: Transformer) -> None:
        transformer_name = get_transformer_name(transformer)
        if transformer_name not in self.transformer_stats:
            self.transformer_stats[transformer_name] = TransformerStats()
        self.transformer_stats[transformer_name].increment_exception_count()

    def increment_transformer_processed_count(self, transformer: Transformer) -> None:
        transformer_name = get_transformer_name(transformer)
        if transformer_name not in self.transformer_stats:
            self.transformer_stats[transformer_name] = TransformerStats()
        self.transformer_stats[transformer_name].increment_processed_count()

    def increment_transformer_filtered_count(self, transformer: Transformer) -> None:
        transformer_name = get_transformer_name(transformer)
        if transformer_name not in self.transformer_stats:
            self.transformer_stats[transformer_name] = TransformerStats()
        self.transformer_stats[transformer_name].increment_filtered_count()

    def increment_action_exception_count(self) -> None:
        self.action_stats.increment_exception_count()

    def increment_action_success_count(self) -> None:
        self.action_stats.increment_success_count()

    def get_started_at(self) -> int:
        return self.started_at

    def get_failed_event_count(self) -> int:
        return self.failed_event_count

    def get_failed_ack_count(self) -> int:
        return self.failed_ack_count

    def get_success_count(self) -> int:
        return self.success_count

    def get_transformer_stats(self, transformer: Transformer) -> TransformerStats:
        transformer_name = get_transformer_name(transformer)
        if transformer_name not in self.transformer_stats:
            self.transformer_stats[transformer_name] = TransformerStats()
        return self.transformer_stats[transformer_name]

    def get_action_stats(self) -> ActionStats:
        return self.action_stats

    def as_string(self) -> str:
        return json.dumps(self.__dict__, indent=4, sort_keys=True)

    def pretty_print_summary(self, name: str) -> None:
        curr_time = int(time() * 1000)
        click.echo()
        click.secho(f"Pipeline Report for {name}", bold=True, fg="blue")
        click.echo()
        click.echo(
            f"Started at: {datetime.datetime.fromtimestamp(self.started_at / 1000.0)} (Local Time)"
        )
        click.echo(f"Duration: {(curr_time - self.started_at) / 1000.0}s")
        click.echo()
        click.secho("Pipeline statistics", bold=True)
        click.echo()
        click.echo(self.as_string())
        click.echo()
        if len(self.transformer_stats.keys()) > 0:
            click.secho("Transformer statistics", bold=True)
            for key in self.transformer_stats:
                click.echo()
                click.echo(f"{key}: {self.transformer_stats[key].as_string()}")
            click.echo()
        click.secho("Action statistics", bold=True)
        click.echo()
        click.echo(self.action_stats.as_string())
        click.echo()



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pathlib
import signal
import sys
import time
from typing import Any, List

import click
from click_default_group import DefaultGroup
from expandvars import UnboundVariable

import datahub_actions._version as actions_version
from datahub.configuration.config_loader import load_config_file
from datahub_actions.pipeline.pipeline import Pipeline
from datahub_actions.pipeline.pipeline_manager import PipelineManager

logger = logging.getLogger(__name__)

# Instantiate a singleton instance of the Pipeline Manager.
pipeline_manager = PipelineManager()


def pipeline_config_to_pipeline(pipeline_config: dict) -> Pipeline:
    logger.debug(
        f"Attempting to create Actions Pipeline using config {pipeline_config.get('name')}"
    )
    try:
        return Pipeline.create(pipeline_config)
    except Exception as e:
        raise Exception(
            f"Failed to instantiate Actions Pipeline using config {pipeline_config.get('name')}: {e}"
        ) from e


@click.group(cls=DefaultGroup, default="run")
def actions() -> None:
    """Execute one or more Actions Pipelines"""
    pass


def load_raw_config_file(config_file: pathlib.Path) -> dict:
    """
    Load a config file as raw YAML/JSON without variable expansion.

    Args:
        config_file: Path to the configuration file

    Returns:
        dict: Raw configuration dictionary

    Raises:
        Exception: If the file cannot be loaded or is invalid YAML/JSON
    """
    try:
        with open(config_file, "r") as f:
            import yaml

            return yaml.safe_load(f)
    except Exception as e:
        raise Exception(
            f"Failed to load raw configuration file {config_file}: {e}"
        ) from e


def is_pipeline_enabled(config: dict) -> bool:
    """
    Check if a pipeline configuration is enabled.

    Args:
        config: Raw configuration dictionary

    Returns:
        bool: True if pipeline is enabled, False otherwise
    """
    enabled = config.get("enabled", True)
    return not (enabled == "false" or enabled is False)


@actions.command(
    name="run",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.option("-c", "--config", required=True, type=str, multiple=True)
@click.option("--debug/--no-debug", default=False)
@click.pass_context
def run(ctx: Any, config: List[str], debug: bool) -> None:
    """Execute one or more Actions Pipelines"""

    logger.info("DataHub Actions version: %s", actions_version.nice_version_name())

    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)

    pipelines: List[Pipeline] = []
    logger.debug("Creating Actions Pipelines...")

    # Phase 1: Initial validation of configs
    valid_configs = []
    for pipeline_config in config:
        pipeline_config_file = pathlib.Path(pipeline_config)
        try:
            # First just load the raw config to check if it's enabled
            raw_config = load_raw_config_file(pipeline_config_file)

            if not is_pipeline_enabled(raw_config):
                logger.warning(
                    f"Skipping pipeline {raw_config.get('name') or pipeline_config} as it is not enabled"
                )
                continue

            valid_configs.append(pipeline_config_file)

        except Exception as e:
            if len(config) == 1:
                raise Exception(
                    f"Failed to load raw configuration file {pipeline_config_file}"
                ) from e
            logger.warning(
                f"Failed to load pipeline configuration! Skipping action config file {pipeline_config_file}...: {e}"
            )

    # Phase 2: Full config loading and pipeline creation
    for pipeline_config_file in valid_configs:
        try:
            # Now load the full config with variable expansion
            pipeline_config_dict = load_config_file(pipeline_config_file)
            pipelines.append(pipeline_config_to_pipeline(pipeline_config_dict))
        except UnboundVariable as e:
            if len(valid_configs) == 1:
                raise Exception(
                    "Failed to load action configuration. Unbound variable(s) provided in config YAML."
                ) from e
            logger.warning(
                f"Failed to resolve variables in config file {pipeline_config_file}...: {e}"
            )
            continue

    # Exit early if no valid pipelines were created
    if not pipelines:
        logger.error(
            f"No valid pipelines were started from {len(config)} config(s). "
            "Check that at least one pipeline is enabled and all required environment variables are set."
        )
        sys.exit(1)

    logger.debug("Starting Actions Pipelines")

    # Start each pipeline
    for p in pipelines:
        pipeline_manager.start_pipeline(p.name, p)
        logger.info(f"Action Pipeline with name '{p.name}' is now running.")

    # Now, run forever only if we have valid pipelines
    while True:
        time.sleep(5)


@actions.command()
def version() -> None:
    """Print version number and exit."""
    click.echo(f"DataHub Actions version: {actions_version.nice_version_name()}")
    click.echo(f"Python version: {sys.version}")


# Handle shutdown signal. (ctrl-c)
def handle_shutdown(signum: int, frame: Any) -> None:
    logger.info("Stopping all running Action Pipelines...")
    pipeline_manager.stop_all()
    sys.exit(1)


signal.signal(signal.SIGINT, handle_shutdown)



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json


# Class that stores running statistics for a single Action.
# TODO: Invocation time tracking.
class ActionStats:
    # The number of exception raised by the Action.
    exception_count: int = 0

    # The number of events that were actually submitted to the Action
    success_count: int = 0

    def increment_exception_count(self) -> None:
        self.exception_count = self.exception_count + 1

    def get_exception_count(self) -> int:
        return self.exception_count

    def increment_success_count(self) -> None:
        self.success_count = self.success_count + 1

    def get_success_count(self) -> int:
        return self.success_count

    def as_string(self) -> str:
        return json.dumps(self.__dict__, indent=4, sort_keys=True)



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Iterable

from datahub.ingestion.api.closeable import Closeable
from datahub_actions.event.event_envelope import EventEnvelope
from datahub_actions.pipeline.pipeline_context import PipelineContext


class EventSource(Closeable, metaclass=ABCMeta):
    """
    The base class for all DataHub Event Sources.

    An Event Source is a producer of DataHub Events which can be acted on using the
    Actions Framework.

    Each Event Source may provide specific semantics, configurations, and processing guarantees.
    Using this interface, the framework can accommodate at-least-once delivery to an individual Action.
    """

    @classmethod
    @abstractmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "EventSource":
        """Factory method to create an instance of an Event Source"""
        pass

    @abstractmethod
    def events(self) -> Iterable[EventEnvelope]:
        """
        Returns an iterable of enveloped events.

        In most cases this should be implemented via a Python generator function which
        can produce a continuous stream of events.
        """

    @abstractmethod
    def ack(self, event: EventEnvelope, processed: bool = True) -> None:
        """
        Acknowledges the processing of an individual event by the Actions Framework
        """



import logging
from typing import Any, Dict, Tuple

logger = logging.getLogger(__name__)


class AckManager:
    """
    An internal Ack manager which hands out fake message ids but ensures we
    track all acks and processing
    """

    def __init__(self) -> None:
        self.batch_id = 0
        self.msg_id = 0
        self.acks: Dict[Tuple[int, int], bool] = {}

    def new_batch(self) -> None:
        self.batch_id += 1
        self.msg_id = 0

    def get_meta(self, event: Any) -> Dict[str, Any]:
        self.msg_id += 1
        self.acks[(self.batch_id, self.msg_id)] = event
        return {"batch_id": self.batch_id, "msg_id": self.msg_id}

    def ack(self, meta: dict, processed: bool) -> None:
        batch_id, msg_id = (meta["batch_id"], meta["msg_id"])
        if processed:
            self.acks.pop((batch_id, msg_id))
        else:
            logger.warning(f"Whoops - we didn't process {meta}")

    def outstanding_acks(self) -> int:
        return len(self.acks)



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from typing import Optional

from pydantic import BaseModel

from datahub_actions.action.action import Action
from datahub_actions.event.event_envelope import EventEnvelope
from datahub_actions.pipeline.pipeline_context import PipelineContext

logger = logging.getLogger(__name__)


class HelloWorldConfig(BaseModel):
    # Whether to print the message in upper case.
    to_upper: Optional[bool] = None


# A basic example of a DataHub action that prints all
# events received to the console.
class HelloWorldAction(Action):
    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "Action":
        action_config = HelloWorldConfig.model_validate(config_dict or {})
        return cls(action_config, ctx)

    def __init__(self, config: HelloWorldConfig, ctx: PipelineContext):
        self.config = config

    def act(self, event: EventEnvelope) -> None:
        print("Hello world! Received event:")
        message = json.dumps(json.loads(event.as_json()), indent=4)
        if self.config.to_upper:
            print(message.upper())
        else:
            print(message)

    def close(self) -> None:
        pass



# Copyright 2021 Acryl Data, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Optional

from datahub_actions.api.action_graph import AcrylDataHubGraph

logger = logging.getLogger(__name__)


class GlossaryTermsResolver:
    def __init__(
        self,
        glossary_entities: Optional[List[str]] = None,
        graph: Optional[AcrylDataHubGraph] = None,
    ):
        self.graph = graph
        self.glossary_term_registry = {}
        self.glossary_node_registry = {}
        if glossary_entities:
            # isolate the terms that don't seem fully specified
            terms_needing_resolution = [
                d
                for d in glossary_entities
                if (not d.startswith("urn:li:glossaryTerm") and d.count("-") != 4)
            ]
            if terms_needing_resolution and not graph:
                raise ValueError(
                    f"Following terms need server-side resolution {terms_needing_resolution} but a DataHub server wasn't provided. Either use fully qualified glossary term ids (e.g. urn:li:glossaryTerm:ec428203-ce86-4db3-985d-5a8ee6df32ba) or provide a datahub_api config in your recipe."
                )
            for term_identifier in terms_needing_resolution:
                self.glossary_term_registry[term_identifier] = (
                    self._resolve_term_id_to_urn(term_identifier)
                )
            nodes_needing_resolution = [
                d
                for d in glossary_entities
                if (not d.startswith("urn:li:glossaryNode") and d.count("-") != 4)
            ]
            if nodes_needing_resolution and not graph:
                raise ValueError(
                    f"Following term groups (glossary nodes) need server-side resolution {nodes_needing_resolution} but a DataHub server wasn't provided. Either use fully qualified glossary term ids (e.g. urn:li:glossaryTerm:ec428203-ce86-4db3-985d-5a8ee6df32ba) or provide a datahub_api config in your recipe."
                )
            for node_identifier in nodes_needing_resolution:
                self.glossary_node_registry[node_identifier] = (
                    self._resolve_node_id_to_urn(node_identifier)
                )

    def _resolve_term_id_to_urn(self, term_identifier: str) -> Optional[str]:
        assert self.graph
        # first try to check if this domain exists by urn
        maybe_term_urn = f"urn:li:glossaryTerm:{term_identifier}"

        if self.graph.graph.exists(maybe_term_urn):
            self.glossary_term_registry[term_identifier] = maybe_term_urn
        else:
            # try to get this term by name
            term_urn = self.graph.get_glossary_term_urn_by_name(term_identifier)
            if term_urn:
                self.glossary_term_registry[term_identifier] = term_urn
            else:
                logger.error(
                    f"Failed to retrieve glossary term id for term {term_identifier}"
                )
                raise ValueError(
                    f"domain {term_identifier} doesn't seem to be provisioned on DataHub. Either provision it first and re-run ingestion, or provide a fully qualified domain id (e.g. urn:li:domain:ec428203-ce86-4db3-985d-5a8ee6df32ba) to skip this check."
                )
        return self.glossary_term_registry.get(term_identifier)

    def _resolve_node_id_to_urn(self, term_identifier: str) -> Optional[str]:
        assert self.graph
        # first try to check if this domain exists by urn
        maybe_term_urn = f"urn:li:glossaryNode:{term_identifier}"

        if self.graph.graph.exists(maybe_term_urn):
            self.glossary_term_registry[term_identifier] = maybe_term_urn
        else:
            # try to get this node by name
            term_urn = self.graph.get_glossary_node_urn_by_name(term_identifier)
            if term_urn:
                self.glossary_node_registry[term_identifier] = term_urn
            else:
                logger.error(
                    f"Failed to retrieve glossary node id for node {term_identifier}"
                )
                raise ValueError(
                    f"domain {term_identifier} doesn't seem to be provisioned on DataHub. Either provision it first and re-run ingestion, or provide a fully qualified domain id (e.g. urn:li:domain:ec428203-ce86-4db3-985d-5a8ee6df32ba) to skip this check."
                )
        return self.glossary_node_registry.get(term_identifier)

    def get_glossary_term_urn(self, term_identifier: str) -> Optional[str]:
        return self.glossary_term_registry.get(
            term_identifier
        ) or self._resolve_term_id_to_urn(term_identifier)

    def get_glossary_node_urn(self, node_identifier: str) -> Optional[str]:
        return self.glossary_node_registry.get(
            node_identifier
        ) or self._resolve_node_id_to_urn(node_identifier)
